import httpx
from starlette.datastructures import MutableHeaders
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.types import ASGIApp
from .umami_api import UmamiRequest, UmamiPayload
from starlette.requests import Request
from starlette.responses import Response
from starlette.background import BackgroundTask
from dataclasses import asdict


async def send_umami_payload(api_endpoint: str, request_payload: UmamiRequest, headers: MutableHeaders,
                             follow_redirects: bool):
    async with httpx.AsyncClient() as client:
        try:
            resp = await client.post(
                api_endpoint,
                json=asdict(request_payload),
                headers=headers,
                follow_redirects=follow_redirects
            )
        except Exception as e:
            # FIXME: How to do exception handling?
            print(f"Error sending umami payload: {e}")


class UmamiMiddleware(BaseHTTPMiddleware):

    def __init__(self,
                 app: ASGIApp,
                 api_endpoint: str,
                 website_id: str,
                 follow_redirects: bool = True,
                 ) -> None:
        super().__init__(app)
        self.app = app
        if not api_endpoint.endswith('/'):
            api_endpoint += '/'
        self.api_endpoint = api_endpoint + 'send'
        self.website_id = website_id
        self.follow_redirects = follow_redirects

    async def dispatch(self, request: Request, call_next) -> Response:
        response = await call_next(request)

        # create umami payload with request data
        umami_request = UmamiRequest(
            payload=UmamiPayload(
                hostname=request.url.hostname,
                language=request.headers.get('Accept-Language', ''),
                referrer=request.headers.get('Referer', ''),
                screen='',
                title='',
                url=request.url.path,
                website=self.website_id,
            )
        )

        # set headers to track IP address correctly
        umami_headers = MutableHeaders()
        # send proper user agent or request won't be registered
        umami_headers['User-Agent'] = request.headers['User-Agent']
        # send IP address of client, not asgi server
        umami_headers['X-Real-IP'] = request.client.host
        umami_headers['X-Forwarded-For'] = request.client.host
        response.background = BackgroundTask(send_umami_payload, self.api_endpoint, umami_request, umami_headers,
                                             self.follow_redirects)

        return response



from fastapi import FastAPI
from starlette.config import Config
from umami_asgi import UmamiMiddleware

config = Config()
UMAMI_API_ENDPOINT = config('UMAMI_API_ENDPOINT')
UMAMI_SITE_ID = config('UMAMI_SITE_ID')

print(f"UMAMI_API_ENDPOINT: {UMAMI_API_ENDPOINT}")

app = FastAPI()

app.add_middleware(UmamiMiddleware, api_endpoint=UMAMI_API_ENDPOINT, website_id=UMAMI_SITE_ID)


@app.get("/")
async def homepage():
    return {"hello": "world"}


@app.get("/feed")
async def feed():
    return {"posts": []}



import pytest
from pytest_httpx import HTTPXMock

from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import PlainTextResponse
from starlette.routing import Route
from starlette.middleware import Middleware
from umami_asgi.middleware import UmamiMiddleware
from starlette.testclient import TestClient
from json import loads


async def feed(request: Request) -> PlainTextResponse:
    return PlainTextResponse("New posts")


async def homepage(request: Request) -> PlainTextResponse:
    return PlainTextResponse("Hello, world")


@pytest.mark.asyncio
async def test_analytics_post(httpx_mock: HTTPXMock):
    httpx_mock.add_response()
    app = Starlette(
        routes=[
            Route("/", endpoint=homepage),
            Route("/feed", endpoint=feed, methods=["GET", "POST"]),
        ],
        middleware=[
            Middleware(UmamiMiddleware, api_endpoint="https://localhost/api", website_id="123456"),
        ],
    )

    with TestClient(app, base_url="http://testserver") as client:
        response = client.get("/")
        assert response.status_code == 200

        # check whether middleware sent post request to mocked umami
        umami_request = httpx_mock.get_request()
        umami_payload = loads(umami_request.content.decode("utf-8"))["payload"]
        assert umami_payload["name"] == "GET"
        assert umami_payload["url"] == "/"
        # check whether middleware set headers correctly to track IP address of client, not asgi server
        assert umami_request.headers["X-Real-IP"] == "testclient"
        assert umami_request.headers["X-Forwarded-For"] == "testclient"

        response = client.post("/feed")
        assert response.status_code == 200

        umami_request = httpx_mock.get_requests()[1]
        umami_payload = loads(umami_request.content.decode("utf-8"))["payload"]
        assert umami_payload["name"] == "POST"
        assert umami_payload["url"] == "/feed"



from dataclasses import dataclass
from typing import Optional


@dataclass
class UmamiPayload:
    """Dataclass for the payload of an Umami request. See https://umami.is/docs/api/sending-stats"""
    hostname: str
    language: str
    referrer: str
    screen: str
    title: str
    url: str
    website: str
    # name: str
    # currently not used -> asdict() outputs "None" if not set, has to be prevented if this field is used
    # data: Optional[dict] = None


@dataclass
class UmamiRequest:
    payload: UmamiPayload
    type: str = "event"  # currently the only type available



from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.responses import JSONResponse
from starlette.config import Config
from starlette.routing import Route

from umami_asgi import UmamiMiddleware

config = Config()
UMAMI_API_ENDPOINT = config('UMAMI_API_ENDPOINT')
UMAMI_SITE_ID = config('UMAMI_SITE_ID')

print(f"UMAMI_API_ENDPOINT: {UMAMI_API_ENDPOINT}")


async def homepage(request):
    return JSONResponse({'hello': 'world'})


async def feed(request):
    return JSONResponse({'posts': []})


app = Starlette(debug=True, routes=[
    Route('/', homepage),
    Route('/feed', feed, methods=['GET', 'POST']),
], middleware=[
    Middleware(UmamiMiddleware, api_endpoint=UMAMI_API_ENDPOINT, website_id=UMAMI_SITE_ID),
])
